/*
 * W3Dmain.c --
 *
 * Procedures to interface the 3D rendering window with the window package
 * for the purposes of window creation, deletion, and modification.
 *
 * Copyright 2003 Open Circuit Design, Inc., for MultiGiG Ltd.
 *
 */

#include <stdio.h>

/* This file is only applicable for OpenGL-enabled graphics;  THREE_D is */
/* defined by the "make config" process and includes all preconditions.	 */
#ifdef THREE_D

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#include <GL/gl.h>
#include <GL/glx.h>

#include "magic/tclmagic.h"
#include "misc/magic.h"
#include "utils/geometry.h"
#include "windows/windows.h"
#include "tiles/tile.h"
#include "utils/hash.h"
#include "database/database.h"
#include "main/main.h"
#include "commands/commands.h"
#include "graphics/wind3d.h"
#include "graphics/graphicsInt.h"
#include "graphics/graphics.h"
#include "grTOGLInt.h"
#include "textio/textio.h"
#include "textio/txcommands.h"
#include "utils/utils.h"
#include "misc/styles.h"
#include "dbwind/dbwtech.h"
#include "dbwind/dbwind.h"
#include "extract/extract.h"
#include "graphics/glyphs.h"
#include "utils/malloc.h"
#include "windows/windInt.h"		/* for access to redisplay pointer */
#include "cif/cif.h"
#include "cif/CIFint.h"		/* access to CIFPlanes, CIFCurStyle, etc. */

extern Display     *grXdpy;		/* X11 display */
extern GLXContext  grXcontext;		/* OpenGL/X11 interface def. */
extern XVisualInfo *grVisualInfo;	/* OpenGL preferred visual */
extern char 	   *MainDisplayType;	/* make sure we're using OpenGL */
extern HashTable   grTOGLWindowTable;
extern int	   grXscrn;
extern int	   grCurFill;

static bool w3dNeedStyle;
static bool w3dIsLocked;
static int  w3dStyle;
static MagWindow *w3dWindow;
extern bool grDriverInformed;

extern void grInformDriver();

global WindClient W3DclientID;

#define glTransYs(n)   (DisplayHeight(grXdpy, grXscrn)-(n))

/* forward declarations */

int W3DCIFredisplay();
int W3Dredisplay();

/* ---------------------------- Commands ------------------------------ */

/* Note that the {x, y, z} specification is primitive and should be	*/
/* changed to something more "natural".					*/

global char *W3DCommandNames[] =
{
       "view x y z	        specify viewpoint angle",
       "scroll x y z	        specify viewpoint position",
       "zoom xy z	specify render volume scale",
	0,
};

int (*w3dCommandProcs[])() =
{
    w3dView,
    w3dScroll,
    w3dZoom,
    0
};

/* ------------------------Low-Level Routines--------------------------------- */

void
w3dLock(w)
    MagWindow *w;
{
    grSimpleLock(w, True);
    w3dSetProjection(w);
}

void
w3dUnlock(w)
    MagWindow *w;
{
    Window wind;
    wind = Tk_WindowId((Tk_Window)w->w_grdata);

    glFlush();
    /* glXSwapBuffers(grXdpy, (GLXDrawable)wind); */
    
    glDisable(GL_COLOR_MATERIAL);
    glDisable(GL_LIGHTING);
    glDisable(GL_DEPTH_TEST);
    glDisable(GL_POLYGON_SMOOTH);

    grSimpleUnlock(w);
}

/* -------------------Low-Level Drawing Routines------------------------------ */

void
w3dFillEdge(r, ztop, zbot)
    Rect *r;
    float ztop;
    float zbot;
{
    float xbot = (float)r->r_xbot;
    float ybot = (float)r->r_ybot;
    float xtop = (float)r->r_xtop;
    float ytop = (float)r->r_ytop;

    glBegin(GL_POLYGON);
    glVertex3f(xbot, ybot, zbot);
    glVertex3f(xbot, ybot, ztop);
    glVertex3f(xtop, ytop, ztop);
    glVertex3f(xtop, ytop, zbot);
    glEnd();
}

void
w3dFillTile(r, zval)
    Rect *r;
    float zval;
{
    float xbot = (float)r->r_xbot;
    float ybot = (float)r->r_ybot;
    float xtop = (float)r->r_xtop;
    float ytop = (float)r->r_ytop;

    glBegin(GL_POLYGON);
    glVertex3f(xbot, ybot, zval);
    glVertex3f(xbot, ytop, zval);
    glVertex3f(xtop, ytop, zval);
    glVertex3f(xtop, ybot, zval);
    glEnd();
}

void
w3dFillXSide(xstart, xend, yval, ztop, zbot)
    float xstart, xend, yval, ztop, zbot;
{
    glBegin(GL_POLYGON);
    glVertex3f(xstart, yval, zbot);
    glVertex3f(xstart, yval, ztop);
    glVertex3f(xend, yval, ztop);
    glVertex3f(xend, yval, zbot);
    glEnd();
}

void
w3dFillYSide(xval, ystart, yend, ztop, zbot)
    float xval, ystart, yend, ztop, zbot;
{
    glBegin(GL_POLYGON);
    glVertex3f(xval, ystart, zbot);
    glVertex3f(xval, ystart, ztop);
    glVertex3f(xval, yend, ztop);
    glVertex3f(xval, yend, zbot);
    glEnd();
}

void
w3dFillOps(trans, tile, ztop, zbot)
    Transform *trans;
    Tile *tile;
    float ztop;
    float zbot;
{
    LinkedRect *tilesegs, *segptr;
    Rect r, r2;
    float xbot, ybot, xtop, ytop;

    r2.r_xbot = LEFT(tile);
    r2.r_ybot = BOTTOM(tile);
    r2.r_xtop = RIGHT(tile);
    r2.r_ytop = TOP(tile);

    GeoTransRect(trans, &r2, &r);

    /* draw tile top and bottom */
    w3dFillTile(&r, ztop);
    w3dFillTile(&r, zbot);

    /* Find tile outline and render top, bottom, and sides */

    if (GrBoxOutline(tile, &tilesegs))
    {
	xbot = (float)r.r_xbot;
	ybot = (float)r.r_ybot;
	xtop = (float)r.r_xtop;
	ytop = (float)r.r_ytop;

	w3dFillXSide(xbot, xtop, ybot, ztop, zbot);
	w3dFillXSide(xbot, xtop, ytop, ztop, zbot);
	w3dFillYSide(xbot, ybot, ytop, ztop, zbot);
	w3dFillYSide(xtop, ybot, ytop, ztop, zbot);
    }
    else
    {
	for (segptr = tilesegs; segptr != NULL; segptr = segptr->r_next)
	{
	    GeoTransRect(trans, &segptr->r_r, &r);
	    w3dFillEdge(&r, ztop, zbot);
	    freeMagic(segptr);
	}
    }
}

void
w3dRenderVolume(tile, trans)
    Tile *tile;
    Transform *trans;
{
    float zbot, ztop;
    float ftop, fthk;
    W3DclientRec *crec;

    crec = (W3DclientRec *)(w3dWindow->w_clientData);

    ExtGetZAxis(tile, &ftop, &fthk);

    /* Negate z-axis values for OpenGL display */
    ztop = -ftop * crec->scale_z;
    zbot = ztop - (fthk * crec->scale_z);

    GR_CHECK_LOCK();
    if (!grDriverInformed)	
	grInformDriver();

    if ((grCurFill == GR_STSOLID) || (grCurFill == GR_STSTIPPLE))
    {
	w3dFillOps(trans, tile, ztop, zbot);
    }

    /* To do:  Outlines and contact crosses OR cif contacts */
    /* To do:  Clip to cross-section plane */
    /* To do:  Deal with non-Manhattan geometry */
}

void
w3dRenderCIF(tile, layer, trans)
    Tile *tile;
    CIFLayer *layer;
    Transform *trans;
{
    float zbot, ztop;
    float ftop, fthk;
    W3DclientRec *crec;

    crec = (W3DclientRec *)(w3dWindow->w_clientData);

    ftop = layer->cl_height;
    fthk = layer->cl_thick;

    /* Negate z-axis values for OpenGL display */
    ztop = -ftop * crec->scale_z;
    zbot = ztop - (fthk * crec->scale_z);

    GR_CHECK_LOCK();
    if (!grDriverInformed)	
	grInformDriver();

    if ((grCurFill == GR_STSOLID) || (grCurFill == GR_STSTIPPLE))
    {
	w3dFillOps(trans, tile, ztop, zbot);
    }

    /* To do:  Outlines */
    /* To do:  Clip to cross-section plane */
    /* To do:  Deal with non-Manhattan geometry */
}

void
w3dClear()
{
    float fr, fg, fb;
    int cidx, lr, lb, lg;

    /* STYLE_TRANSPARENT uses the background color definition,	*/
    /* so we get RGB values from there.				*/

    cidx = GrStyleTable[STYLE_TRANSPARENT].color;
    GrGetColor(cidx, &lr, &lg, &lb);

    fr = ((GLfloat)lr / 255);
    fg = ((GLfloat)lg / 255);
    fb = ((GLfloat)lb / 255);

    glClearColor(fr, fg, fb, 0.0);
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}

/* -------------------High-Level Drawing Routines------------------------------ */

void
w3dSetProjection(w)
    MagWindow *w;
{
    W3DclientRec *crec;
    Window wind;
    GLfloat light_pos[] = { 1.0, 1.0, 1.0, 0.0 };

    crec = (W3DclientRec *) w->w_clientData;
    wind = Tk_WindowId((Tk_Window)w->w_grdata);

    /* Should not mess with surfaceArea---we want the area to be displayed */
    /* in crec, in magic internal coordinates, set by the edit box (else-  */
    /* where).  Set the viewport to maintain 1:1 x:y aspect ratio on the   */
    /* rendered volume.							   */

    glXMakeCurrent(grXdpy, (GLXDrawable)wind, grXcontext);

    if (crec->level > 0)
    {
	glEnable(GL_LINE_SMOOTH);
	glEnable(GL_POLYGON_SMOOTH);
    }
    
    /* Need to look into dealing properly with double-buffered graphics */
    glDrawBuffer(GL_FRONT);
    /* glDrawBuffer(GL_BACK); */

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();

    /* Lighting */  /* (experimental) */

    glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
    glLightfv(GL_LIGHT0, GL_POSITION, light_pos);

    glEnable(GL_COLOR_MATERIAL);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);

    glDepthFunc(GL_LEQUAL);
    glEnable(GL_DEPTH_TEST);

    /* Preserve aspect ratio of 1:1 for internal units on the display */

    glScalef((GLfloat)crec->height / (GLfloat)crec->width, 1.0, 1.0);
    glViewport((GLsizei)0, (GLsizei)0, (GLsizei)crec->width, (GLsizei)crec->height);

    /* Projection matrix manipulations */

    glScalef(crec->scale_xy, crec->scale_xy, crec->prescale_z);

    glRotatef(crec->view_x, 1.0, 0.0, 0.0);
    glRotatef(crec->view_y, 0.0, 1.0, 0.0);
    glRotatef(crec->view_z, 0.0, 0.0, 1.0);

    glTranslatef(crec->trans_x, crec->trans_y, crec->trans_z);
}

/* Magic layer tile painting function */

int
w3dPaintFunc(tile, cxp)
    Tile *tile;			/* Tile to be displayed */
    TreeContext *cxp;		/* From DBTreeSrTiles */
{
    SearchContext *scx = cxp->tc_scx;

    /* Allow display interrupt a la dbwPaintFunc() */
    if (GrDisplayStatus == DISPLAY_BREAK_PENDING)
    {
	GrDisplayStatus = DISPLAY_IN_PROGRESS;
	if (GrEventPendingPtr)
	{
	    if ((*GrEventPendingPtr)())
		sigOnInterrupt(0);
	    else
		SigSetTimer();
	}
    }

    if (!w3dIsLocked)
    {
	w3dLock(w3dWindow);
	w3dIsLocked = True;
    }
    if (w3dNeedStyle)
    {
	GrSetStuff(w3dStyle);
	w3dNeedStyle = False;
    }

    w3dRenderVolume(tile, &scx->scx_trans);
    return 0;			/* keep the search going! */
}

/* CIF layer tile painting function */

int
w3dCIFPaintFunc(tile, arg)
    Tile *tile;			/* Tile to be displayed */
    ClientData *arg;		/* is NULL */
{
    CIFLayer *layer = (CIFLayer *)arg;

    /* Allow display interrupt a la dbwPaintFunc() */
    if (GrDisplayStatus == DISPLAY_BREAK_PENDING)
    {
	GrDisplayStatus = DISPLAY_IN_PROGRESS;
	if (GrEventPendingPtr)
	{
	    if ((*GrEventPendingPtr)())
		sigOnInterrupt(0);
	    else
		SigSetTimer();
	}
    }

    if (!w3dIsLocked)
    {
	w3dLock(w3dWindow);
	w3dIsLocked = True;
    }
    if (w3dNeedStyle)
    {
	/* TxError("CIF layer 0x%x (%s) render style %d\n",
		layer,
		layer->cl_name,
		layer->cl_renderStyle); */
	GrSetStuff(layer->cl_renderStyle + TECHBEGINSTYLES);
	w3dNeedStyle = False;
    }

    w3dRenderCIF(tile, layer, &GeoIdentityTransform);
    return 0;			/* keep the search going! */
}
/* -----------------------Command Procedures------------------------------ */

/*
 * ----------------------------------------------------------------------------
 *
 * w3dView --
 *
 *	Set the viewing angle into the 3D display.
 *
 * ----------------------------------------------------------------------------
 */

Void
w3dView(w, cmd)
    MagWindow *w;
    TxCommand *cmd;
{
    W3DclientRec *crec;
    crec = (W3DclientRec *) w->w_clientData;

    if (cmd->tx_argc == 4)
    {
	/* Pick up x, y, z angles */

	if (StrIsNumeric(cmd->tx_argv[1]) &&
	    StrIsNumeric(cmd->tx_argv[2]) &&
	    StrIsNumeric(cmd->tx_argv[3]))
	{
	    crec->view_x = atof(cmd->tx_argv[1]);
	    crec->view_y = atof(cmd->tx_argv[2]);
	    crec->view_z = atof(cmd->tx_argv[3]);
	}
    }
    else
	TxError("Usage: view angle_x angle_y angle_z\n");
}

/*
 * ----------------------------------------------------------------------------
 *
 * w3dScroll --
 *
 *	Set the viewing position into the 3D display.
 *
 * ----------------------------------------------------------------------------
 */

Void
w3dScroll(w, cmd)
    MagWindow *w;
    TxCommand *cmd;
{
    W3DclientRec *crec;
    crec = (W3DclientRec *) w->w_clientData;

    if (cmd->tx_argc == 4)
    {
	/* Pick up x, y, z translations */

	if (StrIsNumeric(cmd->tx_argv[1]) &&
	    StrIsNumeric(cmd->tx_argv[2]) &&
	    StrIsNumeric(cmd->tx_argv[3]))
	{
	    crec->trans_x = atof(cmd->tx_argv[1]);
	    crec->trans_y = atof(cmd->tx_argv[2]);
	    crec->trans_z = atof(cmd->tx_argv[3]);
	}
    }
    else
	TxError("Usage: scroll pos_x pos_y pos_z\n");
}

/*
 * ----------------------------------------------------------------------------
 *
 * w3dZoom --
 *
 *	Set the viewing scale of the 3D display.
 *
 * ----------------------------------------------------------------------------
 */

Void
w3dZoom(w, cmd)
    MagWindow *w;
    TxCommand *cmd;
{
    W3DclientRec *crec;
    crec = (W3DclientRec *) w->w_clientData;

    if (cmd->tx_argc == 3)
    {
	/* Pick up xy and z scales */

	if (StrIsNumeric(cmd->tx_argv[1]) &&
	    StrIsNumeric(cmd->tx_argv[2]))
	{
	    crec->scale_xy = atof(cmd->tx_argv[1]);
	    crec->scale_z = atof(cmd->tx_argv[2]);
	}
    }
    else
	TxError("Usage: zoom scale_xy scale_z\n");
}

/*
 * ----------------------------------------------------------------------------
 *
 * Set3DDefaults --
 *
 *	Set default values for the viewing parameters (scale, angle,
 *	and translation in 3 dimensions)
 *
 * ----------------------------------------------------------------------------
 */

void
Set3DDefaults(crec)
    W3DclientRec *crec;
{
    crec->trans_x = 0.0;
    crec->trans_y = 0.0;
    crec->trans_z = 0.0;

    crec->scale_xy = 1.0;
    crec->scale_z = 25.0;	/* Height exaggerated by 25x */

    /* The small z-scale value is necessary to keep large layout distances	*/
    /* from exceeding the OpenGL (-1,1) limits of the z-axis.  Distances in Z,	*/
    /* like layer thickness, simply scale up to compensate.			*/

    crec->prescale_z = 0.0001;

    crec->view_x = 0.0;
    crec->view_y = 0.0;
    crec->view_z = 0.0;
}

/* -----------------------Event Procedures------------------------------ */

Void
W3DEventProc(clientData, xevent)
    ClientData clientData;
    XEvent *xevent;
{
    Tk_Window tkwind = (Tk_Window)clientData;
    MagWindow *mw;
    HashEntry *entry;
    Window wind;
    Rect screenRect;
    W3DclientRec *crec = NULL;

    entry = HashLookOnly(&grTOGLWindowTable, tkwind);
    mw = (entry) ? (MagWindow *)HashGetValue(entry) : NULL;
    if (!mw) mw = w3dWindow;
    if (mw)  crec = (W3DclientRec *) mw->w_clientData;

    if (!mw)
    {
	TxError("No MagWindow for event!\n");
	return;
    }

    switch (xevent->type)
    {
	case ButtonPress:
	case ButtonRelease:
	    break;
	case KeyPress: {
	    /* simple stuff, for now */
	    XKeyPressedEvent *KeyPressedEvent = (XKeyPressedEvent *)xevent;
	    KeySym keysym;
	    char inChar[10];
	    int nbytes;
	    bool action = True;

	    nbytes = XLookupString(KeyPressedEvent, inChar, sizeof(inChar),
		&keysym, NULL);

	    switch(keysym)
	    {
		case XK_L:	/* raise rendering level */
		    crec->level++;
		    break;
		case XK_l:	/* lower rendering level */
		    if (crec->level > 0) crec->level--;
		    break;
		case XK_C:	/* do CIF layers */
		    if ((((clientRec *)(W3DclientID))->w_redisplay == W3Dredisplay)
				&& (CIFCurStyle != NULL))
		    {
			((clientRec *)(W3DclientID))->w_redisplay = W3DCIFredisplay;
			crec->scale_xy /= CIFCurStyle->cs_scaleFactor;
			crec->prescale_z /= CIFCurStyle->cs_scaleFactor;
			crec->scale_z *= CIFCurStyle->cs_scaleFactor;
			crec->trans_y *= CIFCurStyle->cs_scaleFactor;
			crec->trans_x *= CIFCurStyle->cs_scaleFactor;
		    }
		    else if (((clientRec *)(W3DclientID))->w_redisplay
				== W3DCIFredisplay)
		    {
			((clientRec *)(W3DclientID))->w_redisplay = W3Dredisplay;
			crec->scale_xy *= CIFCurStyle->cs_scaleFactor;
			crec->prescale_z *= CIFCurStyle->cs_scaleFactor;
			crec->scale_z /= CIFCurStyle->cs_scaleFactor;
			crec->trans_y /= CIFCurStyle->cs_scaleFactor;
			crec->trans_x /= CIFCurStyle->cs_scaleFactor;
		    }

		    break;
		case XK_space:	/* reset to defaults */
		    Set3DDefaults(crec);
		    break;
		case XK_Up:	/* 1/4 screen pans */
		    crec->trans_y -= (0.25 / crec->scale_xy);
		    break;
		case XK_Down:
		    crec->trans_y += (0.25 / crec->scale_xy);
		    break;
		case XK_Left:
		    crec->trans_x += (0.25 / crec->scale_xy);
		    break;
		case XK_Right:
		    crec->trans_x -= (0.25 / crec->scale_xy);
		    break;
		case XK_z:
		    crec->scale_xy *= 2.0;
		    break;
		case XK_Z:
		    crec->scale_xy /= 2.0;
		    break;
		case XK_greater:
		    crec->scale_z *= 2.0;
		    break;
		case XK_less:
		    crec->scale_z /= 2.0;
		    break;
		case XK_1:
		    crec->view_y += 10.0;
		    break;
		case XK_2:
		    crec->view_y -= 10.0;
		    break;
		case XK_3:
		    crec->view_x += 10.0;
		    break;
		case XK_4:
		    crec->view_x -= 10.0;
		    break;
		case XK_5:
		    crec->view_z += 10.0;
		    break;
		case XK_6:
		    crec->view_z -= 10.0;
		    break;
		case XK_7:
		    crec->view_y += 1.0;
		    break;
		case XK_8:
		    crec->view_y -= 1.0;
		    break;
		case XK_9:
		    crec->view_x += 1.0;
		    break;
		case XK_0:
		    crec->view_x -= 1.0;
		    break;
		case XK_minus:
		    crec->view_z += 1.0;
		    break;
		case XK_equal:
		    crec->view_z -= 1.0;
		    break;
		default:
		    action = False;
	    }
	    if (action)
	    {
		screenRect.r_xbot = 0;
		screenRect.r_ybot = 0;
		screenRect.r_xtop = crec->width;
		screenRect.r_ytop = crec->height;
		WindAreaChanged(mw, &screenRect);
		WindUpdate();
	    }
	    }
	    break;
	case ConfigureNotify: {
	    XConfigureEvent *ConfigureEvent = (XConfigureEvent *)xevent;
	    crec->width = ConfigureEvent->width;
	    crec->height = ConfigureEvent->height;
	    screenRect.r_xbot = 0;
	    screenRect.r_ybot = 0;
	    screenRect.r_xtop = ConfigureEvent->width;
	    screenRect.r_ytop = ConfigureEvent->height;
	    WindAreaChanged(mw, &screenRect);
	    WindUpdate();
	    }
	    break;
	case Expose: {
	    XExposeEvent *ExposeEvent = (XExposeEvent *)xevent;
	    screenRect.r_xbot = ExposeEvent->x;
	    screenRect.r_xtop = ExposeEvent->x + ExposeEvent->width;
	    screenRect.r_ytop = mw->w_allArea.r_ytop - ExposeEvent->y;
	    screenRect.r_ybot = mw->w_allArea.r_ytop - (ExposeEvent->y
			+ ExposeEvent->height);
	    WindAreaChanged(mw, &screenRect);
	    WindUpdate();
	    }
	    break;
	case MapNotify:
	case UnmapNotify:
	case DestroyNotify:	/* Do nothing */
	    break;

	default:
	    TxError("Tk Event: Unknown (%d)\n", xevent->type);
	    TxFlush();
	    break;
    }
}

/* -----------------------Client Procedures------------------------------ */

/*
 * ----------------------------------------------------------------------------
 *
 * W3DloadWindow --
 *
 *	Replace the root cell of a window by the specified cell.
 *
 * Results:
 *	True if successful, False if not.  The 3D rendering window is
 *	not allowed to create new cells, only to render existing ones.
 *
 * Side effects:
 *	None.
 *
 * ----------------------------------------------------------------------------
 */

bool
W3DloadWindow(window, name)
    MagWindow *window;
    char *name;
{
    CellDef *newEditDef;
    CellUse *newEditUse;
    Rect loadBox;

    newEditDef = DBCellLookDef(name);
    if (newEditDef == (CellDef *)NULL)
	return False;

    if (!DBCellRead(newEditDef, (char *)NULL, True))
	return False;

    DBReComputeBbox(newEditDef);
    loadBox = newEditDef->cd_bbox;

    /* Attach cell to window */

    newEditUse = DBCellNewUse(newEditDef, (char *) NULL);
    (void) StrDup(&(newEditUse->cu_id), "3D rendered cell");

    window->w_bbox = &(newEditUse->cu_def->cd_bbox);
    return WindLoad(window, W3DclientID, (ClientData)newEditUse, &loadBox);
}

/*
 * ----------------------------------------------------------------------------
 *
 * W3Dcreate --
 *
 * A new window has been created.  Create and initialize the needed 
 * structures.
 *
 * Results:
 *	False if we have too many windows, True otherwise.
 *
 * Side effects:
 *	Initialize the window to be editing the background color.
 *
 * ----------------------------------------------------------------------------
 */

bool
W3Dcreate(window, argc, argv)
    MagWindow *window;
    int argc;
    char *argv[];
{
    W3DclientRec *crec;
    Tk_Window tkwind, tktop;
    Window wind;    
    Colormap colormap;
    HashEntry *entry;
    CellDef *boxDef;
    Rect box;
    bool result;

    /* At least for now, there's only one 3D window allowed.		*/

    if (w3dWindow != NULL)
    {
	TxError("Only one 3D window allowed.\n");
	return False;
    }

    /* The 3D rendering, of course, *only* works with OpenGL, so in an	*/
    /* executable compiled for multiple graphics interfaces, we need	*/
    /* to make sure that our display type is OpenGL.			*/

    if (strncmp(MainDisplayType, "OGL", 3))
    {
	TxError("Display type is \"%s\".  OpenGL is required for the 3D display.\n",	
		MainDisplayType);
	TxError("Please restart magic with option \"-d OGL\".\n");
	return False;
    }
 
    crec = (W3DclientRec *) mallocMagic(sizeof(W3DclientRec));

    /* The MagWindow structure and frameArea indicates the cross-sectional */
    /* area of the layout to be rendered in the 3D display window.	   */

    /* Need to parse the argument list here. . .  At least one argument	*/
    /* should allow the Tk path name to be passed to the routine, as	*/
    /* it is for the standard layout window in the Tk interface.	*/ 

    /* Set surface area, etc. of the MagWindow. . . ? */

    /* Set default values for the client record */

    Set3DDefaults(crec);

    /* Initial window height and width (should be passed as optional	*/
    /* arguments to the specialopen command)				*/

    crec->width = 500;
    crec->height = 500;

    /* Rendering level (0 = coarse & fast, 1 = finer but slower, etc. */

    crec->level = 1;

    window->w_clientData = (ClientData) crec;
    window->w_flags &= ~(WIND_SCROLLABLE | WIND_SCROLLBARS | WIND_CAPTION |
		WIND_BORDER);

    /* Load the current cellDef into the window */

    if (argc > 0)
    {
	result = W3DloadWindow(window, argv[0]);
        if (result == True)
	{
	    boxDef = ((CellUse *)window->w_surfaceID)->cu_def;
	    crec->clipArea = boxDef->cd_bbox;
	}
    }
    else if (ToolGetBox(&boxDef, &box))
    {
	result = W3DloadWindow(window, boxDef->cd_name);
	if (result == True)
	    crec->clipArea = box;
    }
    else
    {
	TxError("No cell exists to render!  Is box in a window?\n");
	return False;
    }

    if (result == False)
    {
	TxError("Cells cannot be created in the 3D window\n");
	return result;
    }

    /* Generate the window. */

    colormap = XCreateColormap(grXdpy, RootWindow(grXdpy, DefaultScreen(grXdpy)),
		grVisualInfo->visual, AllocNone);

    if (!(tktop = Tk_MainWindow(magicinterp))) return False;
	
    tkwind = Tk_CreateWindowFromPath(magicinterp, tktop, ".magic3d", "");

    if (tkwind != 0)
    {
	window->w_grdata = (ClientData) tkwind;
	entry = HashFind(&grTOGLWindowTable, tkwind);
	HashSetValue(entry, window);

	/* visual type must match what we chose in the graphics init proc */
	Tk_SetWindowVisual(tkwind, grVisualInfo->visual, toglCurrent.depth,
		colormap);

	Tk_MapWindow(tkwind);
	
	Tk_GeometryRequest(tkwind, crec->width, crec->height);

	wind = Tk_WindowId(tkwind);
	glXMakeCurrent(grXdpy, (GLXDrawable)wind, grXcontext);

	/* execute any pending Tk events */

	while (Tcl_DoOneEvent(TCL_DONT_WAIT) != 0);

	/* set up our own Tk event handler for the 3d window */

	Tk_CreateEventHandler(tkwind, ExposureMask | StructureNotifyMask |
		ButtonPressMask | ButtonReleaseMask | KeyPressMask,
		(Tk_EventProc *)W3DEventProc, (ClientData)tkwind);

	w3dWindow = window;
	return True;
    }

    TxError("Could not create a new Tk window\n");
    return False;
}

/*
 * ----------------------------------------------------------------------------
 *
 * W3Ddelete --
 *
 * Clean up the data structures before deleting a window.
 *
 * Results:
 *	True if we really want to delete the window, False otherwise.
 *
 * Side effects:
 *	A W3DclientRec is freed.
 *
 * ----------------------------------------------------------------------------
 */

W3Ddelete(window)
    MagWindow *window;
{
    W3DclientRec *cr;
    Tk_Window xw;

    cr = (W3DclientRec *) window->w_clientData;
    xw = (Tk_Window)window->w_grdata;
    w3dWindow = NULL;

    freeMagic((char *) cr);
    Tk_DestroyWindow(xw);
    return (True);
}


/*
 * ----------------------------------------------------------------------------
 *
 * W3Dredisplay --
 *
 * Redisplay a portion of the 3D rendered layout view
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Redisplay is done.
 * ----------------------------------------------------------------------------
 */

int
W3Dredisplay(w, rootArea, clipArea)
    MagWindow *w;		/* The window containing the area. */
    Rect *rootArea;	/* Redisplay area in surface coordinates. */
    Rect *clipArea;	/* An area on the screen to clip to. */
{
    W3DclientRec *crec;
    CellDef *cellDef;
    SearchContext scontext;
    Rect largerArea;
    int i;
    TileTypeBitMask *mask;

    grSimpleLock(w, True);

    crec = (W3DclientRec *) w->w_clientData;
    cellDef = ((CellUse *)w->w_surfaceID)->cu_def;

    if (rootArea != NULL)
        largerArea = *rootArea;
    else
	largerArea = w->w_surfaceArea;

    largerArea.r_xbot--;
    largerArea.r_ybot--;
    largerArea.r_xtop++;
    largerArea.r_ytop++;

    scontext.scx_area = largerArea;
    scontext.scx_use = ((CellUse *)w->w_surfaceID);
    scontext.scx_x = scontext.scx_y = -1;
    scontext.scx_trans = GeoIdentityTransform;

    w3dClear();

    /* follow the same locking procedure as DBWredisplay() */
    w3dUnlock(w);
    w3dIsLocked = False;
    for (i = 0; i < DBWNumStyles; i++)
    {
	/* This should probably be redesigned. . . */
	mask = DBWStyleToTypes(i);
	if (!TTMaskIsZero(mask))
	{
	    w3dStyle = i + TECHBEGINSTYLES;
	    w3dNeedStyle = True;
	    (void) DBTreeSrTiles(&scontext, mask, 0,
			w3dPaintFunc, (ClientData) NULL);
	    if (w3dIsLocked)
	    {
		w3dUnlock(w);
		w3dIsLocked = False;
	    }
	}
    }
}


/*
 * ----------------------------------------------------------------------------
 *
 * W3DCIFredisplay --
 *
 * Redisplay a portion of the 3D rendered layout, in CIF layers.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Redisplay is done.
 * ----------------------------------------------------------------------------
 */

int
W3DCIFredisplay(w, rootArea, clipArea)
    MagWindow *w;		/* The window containing the area. */
    Rect *rootArea;	/* Redisplay area in surface coordinates. */
    Rect *clipArea;	/* An area on the screen to clip to. */
{
    W3DclientRec *crec;
    SearchContext scx;
    CellDef *cellDef;
    Rect largerArea;
    int i;
    TileTypeBitMask *mask;

    grSimpleLock(w, True);

    crec = (W3DclientRec *) w->w_clientData;
    cellDef = ((CellUse *)w->w_surfaceID)->cu_def;

    if (rootArea != NULL)
        largerArea = *rootArea;
    else
	largerArea = w->w_surfaceArea;

    largerArea.r_xbot--;
    largerArea.r_ybot--;
    largerArea.r_xtop++;
    largerArea.r_ytop++;

    /* The following is basically a copy of CIFSeeLayer() */

    CIFErrorDef = cellDef;
    CIFInitCells();
    UndoDisable();
    CIFDummyUse->cu_def = cellDef;
    GEO_EXPAND(&largerArea, CIFCurStyle->cs_radius, &scx.scx_area);
    scx.scx_use = CIFDummyUse;
    scx.scx_trans = GeoIdentityTransform;
    (void) DBTreeSrTiles(&scx, &DBAllButSpaceAndDRCBits, 0,
	cifHierCopyFunc, (ClientData) CIFComponentDef);
    CIFGen(CIFComponentDef, &largerArea, CIFPlanes, &DBAllTypeBits, True, True);
    DBCellClearDef(CIFComponentDef);

    w3dClear();

    /* follow the same locking procedure as DBWredisplay() */
    w3dUnlock(w);
    w3dIsLocked = False;

    for (i = 0; i < CIFCurStyle->cs_nLayers; i++)
    {
	w3dNeedStyle = True;
	DBSrPaintArea((Tile *) NULL, CIFPlanes[i], &TiPlaneRect,
		&CIFSolidBits, w3dCIFPaintFunc,
		(ClientData)(CIFCurStyle->cs_layers[i]));

	if (w3dIsLocked)
	{
	    w3dUnlock(w);
	    w3dIsLocked = False;
	}
    }
}

/*
 * ----------------------------------------------------------------------------
 *
 * W3Dcommand --
 *
 *	This procedure is invoked by the window package whenever a
 *	command is typed while the cursor is over a colormap window.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	TBD
 *
 * ----------------------------------------------------------------------------
 */

W3Dcommand(w, cmd)
    MagWindow *w;
    TxCommand *cmd;
{
    int cmdNum;

    switch (cmd->tx_button)
    {
        case TX_LEFT_BUTTON:
        case TX_RIGHT_BUTTON:
        case TX_MIDDLE_BUTTON:
	    switch (cmd->tx_buttonAction)
	    {
		case TX_BUTTON_UP:
		    /* w3dButtonUp(w, &(cmd->tx_p), cmd->tx_button); */
		    break;
		case TX_BUTTON_DOWN:
		    /* w3dButtonDown(w, &(cmd->tx_p), cmd->tx_button); */
		    break;
	    }
	    break;
        case TX_NO_BUTTON:
	    if (cmd->tx_argc > 0)
	    {
		cmdNum = Lookup(cmd->tx_argv[0], W3DCommandNames);
		ASSERT(cmdNum >= 0, "W3Dcommand");
		(void) (*w3dCommandProcs[cmdNum])(w, cmd);
	    }
	    break;
	default:
	    ASSERT(False, "W3Dcommand");
    }

    UndoNext();
    return (0);
}


/*
 * ----------------------------------------------------------------------------
 *
 * W3Dinit --
 *
 *	Add the 3D rendering window client to the window module.
 *
 * Results:
 *	None.
 *
 * Side effects:
 *	Add ourselves as a client to the window package.
 *
 * ----------------------------------------------------------------------------
 */

W3Dinit()
{
    W3DclientID = WindAddClient("wind3d", W3Dcreate, W3Ddelete,
			W3Dredisplay, W3Dcommand,
			( Void (*)()) NULL, NULL, NULL,
			W3DCommandNames,
			(GrGlyph *) NULL);
}

#endif  /* THREE_D */
